# $Name:  $ 
# $Author: dgp $ 
# $Date: 1997/03/19 21:13:55 $ 
#
# A Tcl 7.4 compatible replacement for the Tcl package command.
#
# This is not a complete implementation of Tcl's own package command in
# versions 7.5 and later.  The known differences are:
#
#	o	The 'package versions' subcommand is not supported.
#	o	Dynamic loading of binary code is not supported
#	o	The entire subcommand must be typed out.  
#		Unambiguous prefixes are not sufficient.
#
# Any software which requires those features should require Tcl 7.5 or later 
# as its requisite support environment.

# So that we can load this package into Tcl 7.5 and later interpreters,
# first make the "package provide" claim using the builtin 'package'
# command.  Also save the existing package database from the built-in
# command for loading into the database of the replacement command.

if {[info commands package] != ""} {

  # There's already a 'package' command defined.  Either this file is
  # being sourced for a second time, or it's being sourced to replace
  # an existing implementation of 'package' (likely the built-in one
  # in Tcl 7.5+).  
  #
  # If this file has already been sourced once, it could be sourced a
  # second time by a 'package require' command only if a 
  # 'package forget Package' has been executed since the 'package require'
  # which previously sourced this file. In that circumstance, a 
  # 'package provide' needs to be executed to restore the forgotten part 
  # of the internal database.
  #
  # If we're replacing an existing implementation of package, then this file
  # is being sourced by a call to 'package require' within that implementation.
  # A call to 'package provide' within that implementation must be made as
  # part of sourcing this file, or that call will return an error,
  # "can't find Package".
  #
  # Both cases are handled by the next two lines of Tcl code.  As it is 
  # coded, it does more than is strictly necessary, but this approach
  # will guarantee that this 'package provide' command stays in sync with 
  # the one at the bottom of this file, since the arguments of both are
  # controlled by CVS.
  package require CvsTools 0
  OMFCVS_PackageProvide {^Name: Package-0_0a0^} pkg

  # Now let's separate the two cases

  if [info exists _omfpkg_package] {

    # This file is being sourced for a second time.  We've done all we need
    # to do, so we can return.  However, another possibile reason this file
    # is being sourced twice is that the file is being explicitly sourced
    # in violation of the package mechanism.  Let's warn against that as we
    # end the sourcing of this file.
    return "Use 'package require Package' to load the Package package."

  } else {

    # There is no internal database structure, so the existing 'package'
    # command is not ours.  It's most likely the built-in one from 
    # Tcl 7.5+ .  Let's remember it, so we can transfer its database
    # contents into our database after we have defined our replacement
    # implementation.
    rename package _OMFPKG_OriginalPackage

  }
}

# proc package
#	See Tcl 7.5 or later documentation on the 'package' command,
#	except the variations noted above.
#
#	This code is closely based on the C implementation of the 
#	package command in Tcl 7.6
#
# Usual proc naming convention not followed so that this is a
# transparent replacement of the Tcl 'package' command.  No namespace
# danger as this is a Tcl core command.
proc package {args} {eval OMFPKG_Package $args}

# This is the real name, which follows conventions for the sake of other
# routines which expect conventions.
proc OMFPKG_Package {command args} {
  global _omfpkg_package _omfpkg_g errorInfo omfpkg_debug
  if $omfpkg_debug {puts "package $command $args"}
  set nargs [llength $args]
  switch $command {
    forget		{ foreach pkgname $args {
			    if [info exists _omfpkg_package($pkgname)] {
			      unset _omfpkg_package($pkgname)
			    } else {
			      return ""
			    }
			  }
			  return ""
			}
    ifneeded		{ 
			  if {$nargs < 2 || $nargs > 3} {
			    return -code error \
"wrong # args: should be \"package ifneeded package version ?script?\""
			  }
			  set pkgname [lindex $args 0]
			  set version [lindex $args 1]
			  _OMFPKG_CheckVersion $version
			  if {![info exists _omfpkg_package($pkgname)]} {
			    if {$nargs == 2} {
			      return ""
			    } else {
			      array set avail {}
			    }
			  } else {
			    if {[llength $_omfpkg_package($pkgname)] > 1} {
			      array set avail $_omfpkg_package($pkgname)
			    } else {
			      return ""
			    }
			    if {$nargs == 2 && ![info exists avail($version)]} {
			      return ""
			    }
			  }
			  if {$nargs == 2} {
			    return $avail($version)
			  }
			  set script [lindex $args 2]
			  set avail($version) $script
			  set _omfpkg_package($pkgname) [array get avail]
			}
    names		{ 
			  if $nargs {
			    return -code error \
"wrong # args: should be \"package names\""
			  }
			  return [array names _omfpkg_package]
			}
    provide		{ 
			  if {$nargs < 1 || $nargs > 2} {
			    return -code error \
"wrong # args: should be \"package provide package ?version?\""
			  }
			  set pkgname [lindex $args 0]
			  if {$nargs == 1} {
			    if [info exists _omfpkg_package($pkgname)] {
			      if {[llength $_omfpkg_package($pkgname)] > 1} {
			        return ""
			      } else {
			        return $_omfpkg_package($pkgname)
			      }
			    } else {
			      return ""
			    }
			  }
			  set version [lindex $args 1]
			  _OMFPKG_CheckVersion $version
			  if {![info exists _omfpkg_package($pkgname)] || \
			       [llength $_omfpkg_package($pkgname)] > 1} {
			    set _omfpkg_package($pkgname) $version
			    return ""
			  } else {
			    if [string compare $_omfpkg_package($pkgname) \
                                               $version] {
			      return -code error \
"conflicting versions provided for package \"$pkgname\":\
$_omfpkg_package($pkgname), then $version"
			    } else {
			      return ""
			    }
			  }
			}
    require		{
			  if {$nargs < 1 || $nargs > 3} {
			    return -code error \
"wrong # args: should be \"package require ?-exact? package ?version?\""
			  }
			  set exact 0
			  if [string compare [lindex $args 0] "-exact"] {
			    if {$nargs == 3} {
			      return -code error \
"wrong # args: should be \"package require ?-exact? package ?version?\""
			    }
			  } else {
			    set args [lrange $args 1 2]
			    set exact 1
			  }
			  set pkgname [lindex $args 0]
			  if {[llength $args] == 2} {
			    set version [lindex $args 1]
			    _OMFPKG_CheckVersion $version
			  } else {
			    set version ""
			  }
			  set foundversion ""
			  for {set pass 1} {1} {incr pass} {
			    if [info exists _omfpkg_package($pkgname)] {
			      if {[llength $_omfpkg_package($pkgname)] > 1} {
			        array set avail $_omfpkg_package($pkgname)
			        set bestmatch ""
			        foreach availversion [array names avail] {
			          if {[string compare $bestmatch ""] &&
			              [_OMFPKG_ComparePkgVersions \
			                 $availversion $bestmatch] <= 0} {
			            continue
			          } 
			          if [string compare $version ""] {
			            set res [_OMFPKG_ComparePkgVersions \
			              $availversion $version sat]
			            if {$res != 0 && $exact} {continue}
			            if {!$sat} {continue}
			          }
			          set bestmatch $availversion
			        }
			        if [string compare $bestmatch ""] {
			          set command $avail($bestmatch)
			          if [catch {eval $command} result] {
			            append errorInfo \
			              "\n\t(\"package ifneeded\" script)"
			            return -code error
			          } else {
			          }
			          continue
			        }
			      } else {
			        set foundversion $_omfpkg_package($pkgname)
			        break
			      }
			    }
			    if {$pass > 1} {break}
			    set command \
			      [list [package unknown] $pkgname $version]
			    if $exact {
			      lappend command -exact
			    }
			    if [catch {eval $command} result] {
			      append errorInfo \
			        "\n\t(\"package unknown\" script)"
			      return -code error
			    }
                          }
			  if $omfpkg_debug {puts "Found version $foundversion"}
			  if [string compare $foundversion ""] {
			    if [string compare $version ""] {
			      set res [_OMFPKG_ComparePkgVersions \
			        $foundversion $version sat]
			      if {($sat && ! $exact) || ($res == 0)} {
			        return $foundversion
			      } else {
			        return -code error \
"version conflict for package \"$pkgname\": have $foundversion, need $version"
			      }
			    } else {
			      return $foundversion
			    }
			  } else {
			    return -code error \
"can't find package $pkgname $version"
			  }

			}
    unknown		{ 
			  switch $nargs {
			    0	{return $_omfpkg_g(unknown)}
			    1	{set _omfpkg_g(unknown) [lindex $args 0]}
			    default	{ return -code error \
"wrong # args: should be \"package unknown ?command?\""
					}
			  }
			}
    vcompare		{
			  if {$nargs != 2} {
			    return -code error \
"wrong # args: should be \"package vcompare version1 version2\""
			  }
			  set version1 [lindex $args 0]
			  set version2 [lindex $args 1]
			  return \
			    [_OMFPKG_ComparePkgVersions $version1 $version2]
			}
    vsatisfies		{
			  if {$nargs != 2} {
			    return -code error \
"wrong # args: should be \"package vsatisfies version1 version2\""
			  }
			  set version1 [lindex $args 0]
			  set version2 [lindex $args 1]
			  _OMFPKG_ComparePkgVersions $version1 $version2 sat
			  return $sat
			}
    default		{return -code error \
"bad option \"$command\": should be forget, ifneeded, names, provide,\n\
 require, unknown, vcompare, or vsatisfies"
			}
  }
}

# proc ComparePkgVersions version1 version2 satref
#	Compares the version numbers version1 and version2.
#	Returns 0 is they are equal, -1 if version1 is lesser,
#	+1 if version2 is lesser.  If satref is non-empty, it
#	names a variable set to 1 if version1 satisfies version2,
#	and 0 otherwise.
proc _OMFPKG_ComparePkgVersions {version1 version2 {satref {}}} {
  _OMFPKG_CheckVersion $version1
  _OMFPKG_CheckVersion $version2
  set major 1; set done1 0; set done2 0;
  while {1} {
    regexp {^([0-9]+)?(\.(.*))?} $version1 match1 num1 foo version1
    if {[string compare $match1 ""] == 0} { set num1 0; set done1 1}
    regexp {^([0-9]+)?(\.(.*))?} $version2 match2 num2 foo version2
    if {[string compare $match2 ""] == 0} { set num2 0; set done2 1}
    if {$num1 != $num2} {break}
    if {$done1 && $done2} {break}
    set major 0
  }
  if [string compare $satref ""] {
    upvar $satref sat
    set sat [expr $num1 == $num2 || ($num1 > $num2 && !$major)]
  }
  if {$num1 > $num2} {
    return 1
  } else {
    if {$num1 < $num2} {
      return -1
    } else {
      return 0
    }
  }
}

# proc CheckVersion version
#	Checks that the string version is a valid Tcl version number.
proc _OMFPKG_CheckVersion {version} {
  if {![regexp {[0-9]+(\.[0-9]+)*} $version]} {
    return -code error "expected version number but got \"$version\""
  }
  return
}

# Define support Tcl procs for package command in Tcl 7.4 
#	These are modified versions of the Tcl procs which support
# 	the package command in Tcl 7.6
#
#	NOTE: tcl_mkIndex is *not* provided.  The intent here is to
#	make software using the package mechanism runnable in a Tcl 7.4
#	interpreter.  Development of that software requires Tcl 7.5 or later

# tclPkgSetup --
# This is a utility procedure use by pkgIndex.tcl files.  It is invoked
# as part of a "package ifneeded" script.  It calls "package provide"
# to indicate that a package is available, then sets entries in the
# auto_index array so that the package's files will be auto-loaded when
# the commands are used.
#
# Arguments:
# dir -			Directory containing all the files for this package.
# pkg -			Name of the package (no version number).
# version -		Version number for the package, such as 2.1.3.
# files -		List of files that constitute the package.  Each
#			element is a sub-list with three elements.  The first
#			is the name of a file relative to $dir, the second is
#			"load" or "source", indicating whether the file is a
#			loadable binary or a script to source, and the third
#			is a list of commands defined by this file.
#
# Usual proc naming conventions not followed here because we are replacing
# a default Tcl proc, and must use the same name.
proc tclPkgSetup {args} {eval OMFPKG_PkgSetup $args}
proc _OMFPKG_PkgSetup {dir pkg version files} {
    global auto_index

    package provide $pkg $version
    foreach fileInfo $files {
	set f [lindex $fileInfo 0]
	set type [lindex $fileInfo 1]
	foreach cmd [lindex $fileInfo 2] {
	    if {$type == "load"} {
# Only add binary package to auto_index if we have 'load' available to
# load them later.  (Probably don't.)
if {[info commands load] == "" } {
              } else {
		set auto_index($cmd) [list load [file join $dir $f] $pkg]
              }
	    } else {
		set auto_index($cmd) [list source [file join $dir $f]]
	    } 
	}
    }
}

# _OMFPKG_PkgUnknown --
# This procedure provides the default for the "package unknown" function.
# It is invoked when a package that's needed can't be found.  It scans
# the auto_path directories and their immediate children looking for
# pkgIndex.tcl files and sources any such files that are found to setup
# the package database.  (On the Macintosh we also search for pkgIndex
# TEXT resources in all files.)
#
# Arguments:
# name -		Name of desired package.  Not used.
# version -		Version of desired package.  Not used.
# exact -		Either "-exact" or omitted.  Not used.

proc _OMFPKG_PkgUnknown {name version {exact {}}} {
    global auto_path env omfpkg_debug
    if $omfpkg_debug {puts "Searching for $name $version ..."}

    if ![info exists auto_path] {
	return
    }
    for {set i [expr [llength $auto_path] - 1]} {$i >= 0} {incr i -1} {
	set dir [lindex $auto_path $i]
	set file [file join $dir pkgIndex.tcl]
	if [file readable $file] {
	    source $file
	}
	foreach file [glob -nocomplain [file join $dir * pkgIndex.tcl]] {
	    if [file readable $file] {
		set dir [file dirname $file]
		source $file
	    }
	}
    }
}

# Modify piece of the auto_loader implementation to be compatible
# with our replacement package command.
#

# auto_reset --
# Destroy all cached information for auto-loading and auto-execution,
# so that the information gets recomputed the next time it's needed.
# Also delete any procedures that are listed in the auto-load index
# except those defined in this file.
#
# Arguments: 
# None.
#
# Usual proc naming conventions not followed here because we are replacing
# a default Tcl proc, and must use the same name.

proc auto_reset {} {OMFPKG_AutoReset}
proc OMFPKG_AutoReset {} {
    global auto_execs auto_index auto_oldpath
    foreach p [info procs] {
	if {[info exists auto_index($p)] && ![string match auto_* $p]
		&& ![string match *OMFPKG_* $p] 
		&& ([lsearch -exact {unknown package tclPkgSetup} $p] < 0)} {
	    rename $p {}
	}
    }
    catch {unset auto_execs}
    catch {unset auto_index}
    catch {unset auto_oldpath}
}

# Default value for tcl_pkgPath : parent of the Tcl library directory
#	Use lappend just in case tcl_pkgPath already exists
lappend tcl_pkgPath [file dirname $tcl_library]
foreach _omfpkg_g(dir) $tcl_pkgPath {
  if {[lsearch -exact $auto_path $_omfpkg_g(dir)] < 0} {
    lappend auto_path $_omfpkg_g(dir)
  }
}
unset _omfpkg_g(dir)

# Pkg needs a bootstrap for the omfpkg_debug variable
if {![info exists omfpkg_debug]} {set omfpkg_debug 0}

# Define the default package unknown handler
package unknown _OMFPKG_PkgUnknown

# At this point, the implementation of the replacement 'package'
# command and its support is complete, and its internal database
# is empty.  Let's fill it with any contents of the database of
# a previous implementation of 'package'.
if {[info commands _OMFPKG_OriginalPackage] != ""} {
  if $omfpkg_debug {
    puts "Transferring package database to Package implementation..."
  }
  foreach _omfpkg_g(p) [_OMFPKG_OriginalPackage names] {
    set _omfpkg_g(v) [_OMFPKG_OriginalPackage provide $_omfpkg_g(p)]
    if {$_omfpkg_g(v)  == ""} {
      foreach _omfpkg_g(v) [_OMFPKG_OriginalPackage versions $_omfpkg_g(p)] {
        package ifneeded $_omfpkg_g(p) $_omfpkg_g(v) \
          [_OMFPKG_OriginalPackage ifneeded $_omfpkg_g(p) $_omfpkg_g(v)]
      }
    } else {
      package provide $_omfpkg_g(p) $_omfpkg_g(v)
    }
  }
  rename _OMFPKG_OriginalPackage {}
  unset _omfpkg_g(p)
  unset _omfpkg_g(v)
}

# Fill in the missing package provide for Tcl
package provide Tcl $tcl_version

# If Tk is in the interpreter, fill in its missing package provide
if [info exists tk_version] {
  package provide Tk $tk_version
}

# We want to use the Cvs package (defines OMFCVS_PackageProvide)
package require CvsTools 0

# Register this file as a package, and define standard global variables
# in case that's useful. (even though it isn't loaded by package require)
OMFCVS_PackageProvide {^Name: Package-0_0a0^} pkg

